import string

KEY_NAMES = ['\t', '\n', '\r', ' ', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', '0', '1',
             '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '_', '`',
             'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u',
             'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~', 'accept', 'add', 'alt', 'altleft', 'altright', 'apps',
             'backspace', 'browserback', 'browserfavorites', 'browserforward', 'browserhome', 'browserrefresh',
             'browsersearch', 'browserstop', 'capslock', 'clear', 'convert', 'ctrl', 'ctrlleft', 'ctrlright', 'decimal',
             'del', 'delete', 'divide', 'down', 'end', 'enter', 'esc', 'escape', 'execute', 'f1', 'f10', 'f11', 'f12',
             'f13', 'f14', 'f15', 'f16', 'f17', 'f18', 'f19', 'f2', 'f20', 'f21', 'f22', 'f23', 'f24', 'f3', 'f4', 'f5',
             'f6', 'f7', 'f8', 'f9', 'final', 'fn', 'hanguel', 'hangul', 'hanja', 'help', 'home', 'insert', 'junja',
             'kana', 'kanji', 'launchapp1', 'launchapp2', 'launchmail', 'launchmediaselect', 'left', 'modechange',
             'multiply', 'nexttrack', 'nonconvert', 'num0', 'num1', 'num2', 'num3', 'num4', 'num5', 'num6', 'num7',
             'num8', 'num9', 'numlock', 'pagedown', 'pageup', 'pause', 'pgdn', 'pgup', 'playpause', 'prevtrack',
             'print', 'printscreen', 'prntscrn', 'prtsc', 'prtscr', 'return', 'right', 'scrolllock', 'select',
             'separator', 'shift', 'shiftleft', 'shiftright', 'sleep', 'space', 'stop', 'subtract', 'tab', 'up',
             'volumedown', 'volumemute', 'volumeup', 'win', 'winleft', 'winright', 'yen', 'command', 'option',
             'optionleft', 'optionright']

KEYBOARD_KEYS = dict([(key, None) for key in KEY_NAMES])

""" Taken from events.h
/System/Library/Frameworks/Carbon.framework/Versions/A/Frameworks/HIToolbox.framework/Versions/A/Headers/Events.h

The *KB dictionaries in a string that can be passed to keyDown(),
keyUp(), or press() into the code used for the OS-specific keyboard function.

They should always be lowercase, and the same keys should be used across all OSes."""

KEYBOARD_KEYS.update({'a': 0x00,  # kVK_ANSI_A
                      's': 0x01,  # kVK_ANSI_S
                      'd': 0x02,  # kVK_ANSI_D
                      'f': 0x03,  # kVK_ANSI_F
                      'h': 0x04,  # kVK_ANSI_H
                      'g': 0x05,  # kVK_ANSI_G
                      'z': 0x06,  # kVK_ANSI_Z
                      'x': 0x07,  # kVK_ANSI_X
                      'c': 0x08,  # kVK_ANSI_C
                      'v': 0x09,  # kVK_ANSI_V
                      'b': 0x0b,  # kVK_ANSI_B
                      'q': 0x0c,  # kVK_ANSI_Q
                      'w': 0x0d,  # kVK_ANSI_W
                      'e': 0x0e,  # kVK_ANSI_E
                      'r': 0x0f,  # kVK_ANSI_R
                      'y': 0x10,  # kVK_ANSI_Y
                      't': 0x11,  # kVK_ANSI_T
                      '1': 0x12,  # kVK_ANSI_1
                      '!': 0x12,  # kVK_ANSI_1
                      '2': 0x13,  # kVK_ANSI_2
                      '@': 0x13,  # kVK_ANSI_2
                      '3': 0x14,  # kVK_ANSI_3
                      '#': 0x14,  # kVK_ANSI_3
                      '4': 0x15,  # kVK_ANSI_4
                      '$': 0x15,  # kVK_ANSI_4
                      '6': 0x16,  # kVK_ANSI_6
                      '^': 0x16,  # kVK_ANSI_6
                      '5': 0x17,  # kVK_ANSI_5
                      '%': 0x17,  # kVK_ANSI_5
                      '=': 0x18,  # kVK_ANSI_Equal
                      '+': 0x18,  # kVK_ANSI_Equal
                      '9': 0x19,  # kVK_ANSI_9
                      '(': 0x19,  # kVK_ANSI_9
                      '7': 0x1a,  # kVK_ANSI_7
                      '&': 0x1a,  # kVK_ANSI_7
                      '-': 0x1b,  # kVK_ANSI_Minus
                      '_': 0x1b,  # kVK_ANSI_Minus
                      '8': 0x1c,  # kVK_ANSI_8
                      '*': 0x1c,  # kVK_ANSI_8
                      '0': 0x1d,  # kVK_ANSI_0
                      ')': 0x1d,  # kVK_ANSI_0
                      ']': 0x1e,  # kVK_ANSI_RightBracket
                      '}': 0x1e,  # kVK_ANSI_RightBracket
                      'o': 0x1f,  # kVK_ANSI_O
                      'u': 0x20,  # kVK_ANSI_U
                      '[': 0x21,  # kVK_ANSI_LeftBracket
                      '{': 0x21,  # kVK_ANSI_LeftBracket
                      'i': 0x22,  # kVK_ANSI_I
                      'p': 0x23,  # kVK_ANSI_P
                      'l': 0x25,  # kVK_ANSI_L
                      'j': 0x26,  # kVK_ANSI_J
                      "'": 0x27,  # kVK_ANSI_Quote
                      '"': 0x27,  # kVK_ANSI_Quote
                      'k': 0x28,  # kVK_ANSI_K
                      ';': 0x29,  # kVK_ANSI_Semicolon
                      ':': 0x29,  # kVK_ANSI_Semicolon
                      '\\': 0x2a,  # kVK_ANSI_Backslash
                      '|': 0x2a,  # kVK_ANSI_Backslash
                      ',': 0x2b,  # kVK_ANSI_Comma
                      '<': 0x2b,  # kVK_ANSI_Comma
                      '/': 0x2c,  # kVK_ANSI_Slash
                      '?': 0x2c,  # kVK_ANSI_Slash
                      'n': 0x2d,  # kVK_ANSI_N
                      'm': 0x2e,  # kVK_ANSI_M
                      '.': 0x2f,  # kVK_ANSI_Period
                      '>': 0x2f,  # kVK_ANSI_Period
                      '`': 0x32,  # kVK_ANSI_Grave
                      '~': 0x32,  # kVK_ANSI_Grave
                      ' ': 0x31,  # kVK_Space
                      'space': 0x31, '\r': 0x24,  # kVK_Return
                      '\n': 0x24,  # kVK_Return
                      'enter': 0x24,  # kVK_Return
                      'return': 0x24,  # kVK_Return
                      '\t': 0x30,  # kVK_Tab
                      'tab': 0x30,  # kVK_Tab
                      'backspace': 0x33,  # kVK_Delete, which is "Backspace" on OS X.
                      '\b': 0x33,  # kVK_Delete, which is "Backspace" on OS X.
                      'esc': 0x35,  # kVK_Escape
                      'escape': 0x35,  # kVK_Escape
                      'command': 0x37,  # kVK_Command
                      'shift': 0x38,  # kVK_Shift
                      'shiftleft': 0x38,  # kVK_Shift
                      'capslock': 0x39,  # kVK_CapsLock
                      'option': 0x3a,  # kVK_Option
                      'optionleft': 0x3a,  # kVK_Option
                      'alt': 0x3a,  # kVK_Option
                      'altleft': 0x3a,  # kVK_Option
                      'ctrl': 0x3b,  # kVK_Control
                      'ctrlleft': 0x3b,  # kVK_Control
                      'shiftright': 0x3c,  # kVK_RightShift
                      'optionright': 0x3d,  # kVK_RightOption
                      'ctrlright': 0x3e,  # kVK_RightControl
                      'fn': 0x3f,  # kVK_Function
                      'f17': 0x40,  # kVK_F17
                      'volumeup': 0x48,  # kVK_VolumeUp
                      'volumedown': 0x49,  # kVK_VolumeDown
                      'volumemute': 0x4a,  # kVK_Mute
                      'f18': 0x4f,  # kVK_F18
                      'f19': 0x50,  # kVK_F19
                      'f20': 0x5a,  # kVK_F20
                      'f5': 0x60,  # kVK_F5
                      'f6': 0x61,  # kVK_F6
                      'f7': 0x62,  # kVK_F7
                      'f3': 0x63,  # kVK_F3
                      'f8': 0x64,  # kVK_F8
                      'f9': 0x65,  # kVK_F9
                      'f11': 0x67,  # kVK_F11
                      'f13': 0x69,  # kVK_F13
                      'f16': 0x6a,  # kVK_F16
                      'f14': 0x6b,  # kVK_F14
                      'f10': 0x6d,  # kVK_F10
                      'f12': 0x6f,  # kVK_F12
                      'f15': 0x71,  # kVK_F15
                      'help': 0x72,  # kVK_Help
                      'home': 0x73,  # kVK_Home
                      'pageup': 0x74,  # kVK_PageUp
                      'pgup': 0x74,  # kVK_PageUp
                      'del': 0x75,  # kVK_ForwardDelete
                      'delete': 0x75,  # kVK_ForwardDelete
                      'f4': 0x76,  # kVK_F4
                      'end': 0x77,  # kVK_End
                      'f2': 0x78,  # kVK_F2
                      'pagedown': 0x79,  # kVK_PageDown
                      'pgdn': 0x79,  # kVK_PageDown
                      'f1': 0x7a,  # kVK_F1
                      'left': 0x7b,  # kVK_LeftArrow
                      'right': 0x7c,  # kVK_RightArrow
                      'down': 0x7d,  # kVK_DownArrow
                      'up': 0x7e,  # kVK_UpArrow
                      })

for char in string.ascii_lowercase:  # Add mappings for uppercase letters.
    KEYBOARD_KEYS[char.upper()] = KEYBOARD_KEYS[char]

# Taken from ev_keymap.h
# http://www.opensource.apple.com/source/IOHIDFamily/IOHIDFamily-86.1/IOHIDSystem/IOKit/hidsystem/ev_keymap.h
SPECIAL_KEYS = {'KEYTYPE_SOUND_UP': 0, 'KEYTYPE_SOUND_DOWN': 1, 'KEYTYPE_BRIGHTNESS_UP': 2,
                'KEYTYPE_BRIGHTNESS_DOWN': 3, 'KEYTYPE_CAPS_LOCK': 4, 'KEYTYPE_HELP': 5, 'POWER_KEY': 6,
                'KEYTYPE_MUTE': 7, 'UP_ARROW_KEY': 8, 'DOWN_ARROW_KEY': 9, 'KEYTYPE_NUM_LOCK': 10,
                'KEYTYPE_CONTRAST_UP': 11, 'KEYTYPE_CONTRAST_DOWN': 12, 'KEYTYPE_LAUNCH_PANEL': 13, 'KEYTYPE_EJECT': 14,
                'KEYTYPE_VIDMIRROR': 15, 'KEYTYPE_PLAY': 16, 'KEYTYPE_NEXT': 17, 'KEYTYPE_PREVIOUS': 18,
                'KEYTYPE_FAST': 19, 'KEYTYPE_REWIND': 20, 'KEYTYPE_ILLUMINATION_UP': 21,
                'KEYTYPE_ILLUMINATION_DOWN': 22, 'KEYTYPE_ILLUMINATION_TOGGLE': 23}



import datetime
import hashlib
import math
import os
import shutil
import zipfile
from pathlib import Path
from pwd import getpwuid
from typing import Optional, List, Tuple


class FileManager:
    permissions = {'all': 0o777, 'default': 0o755, 'read_only': 0o444, 'write_only': 0o222, 'no_access': 0o000}

    def __init__(self, executor):
        self.executor = executor

    def clear_quarantine_attribute(self, path: str):
        self.executor.sudo(f'xattr -r -d com.apple.quarantine "{path}"')

    def read_metadata(self, path: str, key: Optional[str] = None):
        if key is None:
            return self.executor.get_output(f'mdls "{path}"')
        return self.executor.get_output(f'mdls -raw -name {key} "{path}"')

    def touch(self, path):
        """
        Sets the modification and access times of files.
        If any file does not exist, it is created with default permissions.
        """
        self.executor.execute(f'touch "{path}"')

    def dump_text_to_file(self, text: str, file_: str):
        command = f'printf "{text}" > "{file_}"'
        if self.executor.execute(command).returncode != 0:
            self.executor.sudo(command)

    def make_file(self, path: str, size: str = '8m', sudo: bool = False):
        """Create a file with size `size`[b|k|m|g] in `path`."""
        execute = self.executor.sudo if sudo else self.executor.execute
        return execute(f'mkfile {size} "{path}"')

    def make_dir(self, path: str, sudo: bool = False):
        execute = self.executor.sudo if sudo else self.executor.execute
        return execute(f'mkdir -p "{path}"')

    def last_modified__timedelta(self, path: str):
        timedelta = datetime.datetime.now() - datetime.datetime.fromtimestamp(self.last_modified(path))
        return int(timedelta.total_seconds())

    @staticmethod
    def last_modified(path: str):
        return os.path.getmtime(path)

    @staticmethod
    def is_path_exists(path: str) -> bool:
        return os.path.exists(path)

    @staticmethod
    def is_dir(path: str) -> bool:
        return os.path.isdir(path)

    @staticmethod
    def is_file(path: str) -> bool:
        return os.path.isfile(path)

    @staticmethod
    def read_file(source: str) -> str:
        with open(source, 'r', encoding="ISO-8859-1") as f:
            return f.read()

    @staticmethod
    def write_to_file(source: str, data: str) -> None:
        with open(source, 'w') as f:
            f.write(data)

    def move_tree(self, source, destination):
        """Move `source` file to `destination`."""
        self.copy_tree(source, destination)
        self.remove_tree(source)

    def remove_content_from(self, content, directory) -> None:
        """Remove `content` from `directory`"""
        if not content:
            return
        for item in content:
            self.remove_tree(os.path.join(directory, item))

    @staticmethod
    def rename_file(default_file_path, updated_file_path):
        return os.rename(default_file_path, updated_file_path)

    def clear_directory(self, directory: str, pattern: str = '*', sudo: bool = False):  # Clear all files, by default.
        return (self.executor.sudo if sudo else self.executor.execute)(f'rm -rf {directory}/{pattern}')

    def remove_tree(self, tree_path):
        """Delete a directory tree or a file."""
        if not os.path.exists(tree_path):
            return
        try:
            (shutil.rmtree if os.path.isdir(tree_path) else os.unlink)(tree_path)
        except (OSError, PermissionError):
            self.executor.sudo(f'rm -r "{tree_path}"')

    def copy_file(self, source: str, destination: str):
        self.executor.execute(f'cp "{source}" "{destination}"')

    def move_file(self, source: str, destination: str):
        self.executor.execute(f'mv "{source}" "{destination}"')

    def copy_tree(self, source: str, destination: str):
        """Copy a directory tree or folder's content to the destination."""
        self.makedir(destination)
        execute = self.executor.sudo if self.__are_elevated_rights_needed(destination) else self.executor.execute
        execute(f'rsync -q -ahP "{source}" "{destination}"')

    def __are_elevated_rights_needed(self, destination: str):
        result = False
        if self.get_file_owner(destination) == 'root':
            result = True
        if destination.startswith('/var') or destination.startswith('/private') or destination.startswith('/Library'):
            result = True
        return result

    def get_size(self, file_path) -> int:
        """Total size, in KB"""
        if Path(file_path).exists():
            return int(self.executor.get_output(f'du -hk "{file_path}"').split()[0].strip())
        raise FileNotFoundError(file_path)

    def folder_newest_file(self, folder: str) -> str:
        if Path(folder).is_dir():
            return max(self.folder_content(folder), key=os.path.getctime)
        raise NotADirectoryError(folder)

    @staticmethod
    def folder_content(folder: str) -> List[str]:
        if Path(folder).is_dir():
            return [path_.as_posix() for path_ in Path(folder).glob('*') if not path_.name.startswith('.')]
        raise NotADirectoryError(folder)

    def get_bundle_content_owners(self, path: str) -> List[Tuple[str, str, str]]:
        raw = (line for line in self.executor.get_output(f'ls -R -l "{path}"').splitlines())
        payload = (line.split() for line in raw if line and not (line.startswith('total') or line.startswith('/')))
        return [(line[-1], line[2], line[3]) for line in payload]

    @staticmethod
    def get_file_owner(file_path):
        try:
            return getpwuid(os.stat(file_path).st_uid).pw_name
        except OSError as e:
            if e.errno == 2:
                return 'missing'
            else:
                raise

    @staticmethod
    def unzip_file(src, dst):
        with zipfile.ZipFile(src) as z:
            z.extractall(dst)

    def unzip_app(self, src, dst):
        if not os.path.exists(dst):
            os.makedirs(dst, exist_ok=True)
        self.executor.execute(f'unzip -qq -o "{src}" -d "{dst}"')

    @staticmethod
    def zip_file(src, dst):
        relroot = os.path.abspath(os.path.join(src, os.pardir))
        with zipfile.ZipFile(dst, 'w', zipfile.ZIP_DEFLATED) as zip_:
            for root, dirs, files in os.walk(src):
                zip_.write(root, os.path.relpath(root, relroot))
                for file_ in files:
                    filename = os.path.join(root, file_)
                    if os.path.isfile(filename):
                        arcname = os.path.join(os.path.relpath(root, relroot), file_)
                        zip_.write(filename, arcname)

    def makedir(self, destination):
        if not os.path.exists(destination):
            if self.executor.execute(f'mkdir -p "{destination}"').returncode != 0:
                self.executor.sudo(f'mkdir -p "{destination}"')

    def lock_file(self, path: str):
        if isinstance(path, Path):
            path = path.as_posix()
        return self.executor.sudo(f'chflags uchg "{path}"')

    def change_file_owner(self, file: str, owner: str):
        return self.executor.sudo(f'chown {owner} "{file}"')

    def change_file_permissions(self, path: str, permissions, sudo: bool = False):
        """Change file permissions."""
        (self.executor.sudo if sudo else self.executor.execute)(f'chmod {permissions} "{path}"')

    @staticmethod
    def change_files_permissions(folder, permissions):
        """Change file permissions in `folder`."""
        for root, subfolders, files in os.walk(folder):
            [os.chmod(os.path.join(root, file_), permissions) for file_ in files]

    @staticmethod
    def convert_file_size_to_human_readable(size):
        if size == '':
            result = 0
        else:
            size = float(size)
            if size == 0:
                return '0B'
            size_name = ('bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
            i = int(math.floor(math.log(size, 1000)))
            p = math.pow(1000, i)
            s = round(size / p, 2)
            result = '{} {}'.format(s, size_name[i])
        return result

    @staticmethod
    def convert_file_size_to_bytes(string):
        _size, _unit = string.strip().lower().replace('zero', '0').replace(',', '').split()
        return int(float(_size) * 1000 ** ('bytes', 'kb', 'mb', 'gb', 'tb').index(_unit))

    def find_duplicates_number(self, folders):
        """Return duplicates in format {filename:number of duplicates}"""
        if isinstance(folders, str):
            folders = [folders, ]
        hashes: dict = dict()
        duplicates: dict = dict()
        for folder in folders:
            for dirpath, dirnames, files_list in os.walk(folder):
                for filename in files_list:
                    if not filename.startswith('.'):
                        full_path = os.path.join(dirpath, filename)
                        if not os.path.islink(full_path) and os.path.exists(full_path):
                            file_id = self._generate_md5_checksum(full_path)
                            duplicate = hashes.get(file_id)
                            if duplicate:
                                counter = duplicates.get(filename, 1)
                                duplicates[filename] = counter + 1
                                hashes[file_id].append(full_path)
                            else:
                                hashes[file_id] = [full_path]
        return duplicates

    @staticmethod
    def _generate_md5_checksum(file_name, chunk_size=1024):
        """Calculate md5 checksum of `file_name`."""
        hash_ = hashlib.md5()
        with open(file_name, 'rb') as f:
            chunk = f.read(chunk_size)
            while chunk:
                hash_.update(chunk)
                chunk = f.read(chunk_size)
        return hash_.hexdigest()



import logging
import sqlite3
from typing import Optional


class SQLite3DataBaseManager:
    """SQLite database manager."""

    def __init__(self, db_location: str):
        self.db = db_location
        self.connection: Optional[sqlite3.Connection] = None
        self.cursor: Optional[sqlite3.Cursor] = None
        self.connect(self.db)

    def connect(self, database) -> None:
        try:
            self.connection = sqlite3.connect(database)
            self.cursor = self.connection.cursor()
        except sqlite3.OperationalError:
            logging.warning(f'Could not connect to TCC database: {self.db}')

    def disconnect(self) -> None:
        try:
            self.cursor.close()
            self.connection.close()
        except (AttributeError, sqlite3.ProgrammingError):
            logging.warning(f'Could not disconnect from TCC database: {self.db}')


class DataBaseManager:
    """Database manager factory."""

    def __new__(cls, db_location: str, db_type: str = 'sqlite3'):
        databases = {
            'sqlite3': SQLite3DataBaseManager,
        }
        database = databases.get(db_type)
        if not database:
            raise LookupError(f'Database: "{db_type}" is not found')
        return database(db_location)



import time
from datetime import datetime, timedelta
from typing import ClassVar

from macuitest.lib.operating_system.shell_executor import ShellExecutor


class TimeManager:
    time_format: ClassVar[str] = '%Y.%m.%d-%H:%M:%S'

    def __init__(self, executor: ShellExecutor, ntp_server: str = 'ua.pool.ntp.org'):
        self.ntp_server = ntp_server
        self.executor = executor

    def move_time(self, days: int = 0, hours: int = 0, minutes: int = 5, seconds: int = 0) -> None:
        self.disable_using_network_time()
        start_point: datetime = datetime.now()
        delta = (start_point + timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds + 1)) \
            .strftime(self.time_format)
        self.executor.sudo(f'date -f {self.time_format} {delta}')
        time.sleep(3)
        elapsed, target_time_shift = (datetime.now() - start_point).seconds, hours * 3600 + minutes * 60 + seconds
        assert elapsed >= target_time_shift, f'Elapsed: {elapsed}, target: {target_time_shift}'

    def reset_time(self, ntp_server: str = 'time.apple.com'):
        """Enable using network time and get it by Simple Network Time Protocol from an NTP server. """
        if self.executor.sudo(f'sntp -sS {ntp_server}').returncode != 0:
            self.executor.sudo(f'sntp -s {ntp_server}')
        self.enable_using_network_time()

    @property
    def is_network_time_used(self) -> bool:
        result = self.executor.sudo_get_output('systemsetup -getusingnetworktime').split(':')[-1].strip().lower()
        return {'on': True, 'off': False}.get(result)

    def enable_using_network_time(self):
        self.executor.sudo('systemsetup -setusingnetworktime on')

    def disable_using_network_time(self):
        self.executor.sudo('systemsetup -setusingnetworktime off')



import time
from collections import namedtuple
from pathlib import Path
from typing import List, Union

from macuitest.lib.applescript_lib.applescript_wrapper import as_wrapper
from macuitest.lib.core import wait_condition
from macuitest.lib.operating_system.shell_executor import ShellExecutor


class ServiceManager:
    Service = namedtuple('Service', ['id', 'is_system_service'])

    def __init__(self, executor: ShellExecutor):
        self.executor = executor

    def launch_application_by_bundle_id(self, bundle_id: str, process_name: str) -> bool:
        self.executor.execute(f'open -b "{bundle_id}"')
        return self.wait_process_appeared(process_name)

    def launch_application_by_name(self, app_name: str) -> bool:
        self.executor.execute(f'open -a "{app_name}"')
        return self.wait_process_appeared(app_name)

    def launch_application_by_path(self, path: str) -> bool:
        self.executor.execute(f'open "{path}"')
        return self.wait_process_appeared(Path(path).stem)

    def force_kill_process(self, process: str, sudo: bool = False):
        _execute = self.executor.sudo if sudo else self.executor.execute
        return _execute(f'pkill -9 {process}')

    def kill_process(self, process, sudo=False, pause: Union[int, float] = 1):
        time.sleep(pause)
        _execute = self.executor.sudo if sudo else self.executor.execute
        _execute(f'killall -9 "{process}"')
        return self.wait_process_disappeared(process)

    def wait_process_appeared(self, process_name, timeout=10):
        return wait_condition(lambda: process_name in self.processes_list, timeout=timeout)

    def wait_process_disappeared(self, process_name, timeout=10):
        return wait_condition(lambda: process_name not in self.processes_list, timeout=timeout)

    @property
    def processes_list(self) -> List[str]:
        """Return list of system processes."""
        return as_wrapper.tell_sys_events('return name of every process')

    @staticmethod
    def add_login_item(app_name, is_hidden=False):
        app_path = '/System/Applications' if Path(f'/System/Applications/{app_name}.app').exists() else '/Applications'
        _cmd = 'make login item at end with properties {path:"%s/%s.app", hidden:%s}' % (app_path, app_name, is_hidden)
        return as_wrapper.tell_sys_events(_cmd)

    @staticmethod
    def add_login_item_by_path(path_):
        return as_wrapper.tell_sys_events('make login item at end with properties {path:"%s"}' % path_)

    def clear_login_items(self):
        return [self.remove_login_item(login_item) for login_item in self.login_items]

    @staticmethod
    def remove_login_item(login_item: str):
        return as_wrapper.tell_sys_events(f'delete login item "{login_item}"')

    @property
    def login_items(self, hidden=False):
        _cmd = 'get the name of every login item'
        if hidden:
            _cmd += ' whose hidden is true'
        return as_wrapper.tell_sys_events(_cmd)

    @property
    def cron_jobs(self) -> str:
        return self.executor.get_output('crontab -l')

    def remove_cron_jobs(self) -> int:
        return self.executor.execute('crontab -r')

    def load_service(self, service: Service) -> int:
        return self._manipulate_service('load', service)

    def unload_service(self, service: Service) -> int:
        return self._manipulate_service('unload', service)

    def remove_service(self, service: Service) -> int:
        return self._manipulate_service('remove', service)

    def _manipulate_service(self, action: str, service: Service) -> int:
        _execute = self.executor.sudo if service.is_system_service else self.executor.execute
        flag = ' ' if action == 'remove' else ' -wF '
        return _execute(f'launchctl {action}{flag}{service.id}')

    def did_user_service_quit(self, service: str) -> bool:
        return wait_condition(lambda: service not in self.user_services)

    def did_user_service_launch(self, service: str) -> bool:
        return wait_condition(lambda: service in self.user_services)

    def did_system_service_launch(self, service: str) -> bool:
        return wait_condition(lambda: service in self.system_services)

    @property
    def user_services(self) -> List[str]:
        return self.__get_services(domain='user')

    @property
    def system_services(self) -> List[str]:
        return self.__get_services(domain='system')

    def __get_services(self, domain: str = 'user') -> List[str]:
        _execute = self.executor.get_output if domain == 'user' else self.executor.sudo_get_output
        return [item.split('\t')[-1] for item in _execute('launchctl list').splitlines()]

    def unload_kext(self, kext_id) -> int:
        return self.executor.sudo(f'kextunload -b {kext_id}')

    def get_process_cpu_usage(self, process_name: str) -> float:
        pid = wait_condition(lambda: self.executor.get_output(f'pgrep {process_name}$'), timeout=3)
        if pid:
            return float(self.executor.get_output(f'ps -p {pid} -o %cpu').split('%CPU')[1].strip()).__round__(2)
        raise LookupError(f'Process "{process_name}" not found')



import os
from typing import Optional


class DiskMountFailed(Exception):
    pass


class DiskUtil:
    def __init__(self, executor):
        self.executor = executor

    def mount_disk_image(self, disk_image: str, mountpoint: Optional[str] = None) -> None:
        """Mount `disk_image`."""
        if not os.path.exists(disk_image):
            raise FileNotFoundError(f'File: {disk_image} does not exist')
        cmd = f'hdiutil mount {disk_image} -mountpoint {mountpoint}' if mountpoint else f'hdiutil mount {disk_image}'
        if self.executor.execute(cmd).returncode != 0:
            raise DiskMountFailed(f'Disk mount failed: {disk_image}')

    def detach_disk_image(self, mountpoint: str) -> None:
        """Unmount attached disk image."""
        self.executor.execute(f'hdiutil detach "{mountpoint}" -force')



import time
from typing import Tuple, ClassVar

from macuitest.config.constants import DNSMapping


class Hosts:
    hosts_file: ClassVar[str] = '/etc/hosts'

    def __init__(self, executor, file_manager):
        self.executor = executor
        self.file_manager = file_manager

    def change_mapping(self, mapping: DNSMapping):
        """ Configure local domain resolution for a certain resource. """
        self.file_manager.change_file_permissions(path=self.hosts_file, permissions='777', sudo=True)
        with open(self.hosts_file) as hosts:
            hosts_config = hosts.read().rstrip().splitlines()
        hosts_config.append(mapping)
        updated = '\n'.join(hosts_config)
        with open(self.hosts_file, 'w') as hosts:
            hosts.write(f'{updated}\n')
        self.file_manager.change_file_permissions(path=self.hosts_file, permissions='755', sudo=True)

    def reset(self, known_mappings: Tuple[DNSMapping, ...], disable_pause: bool = False):
        """ Reset /etc/hosts by removing certain/all known mappings. """
        self.file_manager.change_file_permissions(path=self.hosts_file, permissions='777', sudo=True)
        with open(self.hosts_file, 'r') as hosts:
            hosts_config = hosts.read().rstrip().splitlines()
        updated = '\n'.join((line for line in hosts_config if line not in known_mappings))
        with open(self.hosts_file, 'w') as hosts:
            hosts.write(f'{updated}\n')
        self.file_manager.change_file_permissions(path=self.hosts_file, permissions='755', sudo=True)
        self.flush_dns_cache(disable_pause=disable_pause)

    def flush_dns_cache(self, disable_pause: bool = False):
        """ Flush DNS cache to remove stored website location details. """
        time.sleep(0 if disable_pause else 2)
        self.executor.execute('dscacheutil -flushcache')
        time.sleep(0 if disable_pause else 1)
        self.executor.sudo('killall -HUP mDNSResponder')
        time.sleep(0 if disable_pause else 5)  # For macOS to catch up.



class WrongApplicationSignature(Exception):
    pass


class SecurityManager:
    """Wrapper around the Command line interface to keychains and Security framework."""
    _cli: str = 'security'

    def __init__(self, executor):
        self.executor = executor

    def enable_gatekeeper(self):
        """Allow opening apps from identified developers only."""
        self.executor.sudo('spctl --master-enable')

    def disable_gatekeeper(self):
        """Allow opening apps from unidentified developers."""
        self.executor.sudo('spctl --master-disable')

    def check_package_signature(self, package_path: str) -> int:
        return self.executor.execute(f'pkgutil --check-signature "{package_path}"').returncode

    def check_application_signature(self, application_path):
        output = self.executor.sudo_get_output(f'codesign -v --deep {application_path}')
        if output:
            error_message = f'{application_path} is not sign correctly: {output}'
            raise WrongApplicationSignature(error_message)

    def remove_keychain_entry(self, service_name: str, flag: str = 's') -> None:
        """Remove a keychain by a service name."""
        while self.find_generic_password(service_name, flag) == 0:
            self.delete_generic_password(service_name, flag)

    def find_generic_password(self, service_name: str, flag: str = 's') -> int:
        """Find a password in login keychain by a service name."""
        return self.executor.execute(
            f'{self._cli} find-generic-password -{flag} {service_name} > /dev/null 2>&1').returncode

    def delete_generic_password(self, service_name: str, flag: str = 's') -> int:
        """Delete a password in login keychain by a service name."""
        return self.executor.execute(
            f'{self._cli} delete-generic-password -{flag} {service_name} > /dev/null 2>&1').returncode

    def reset_authorization_database(self, search_pattern: str = 'Auth') -> int:
        """Clear the system.login.console authorization rights."""
        result = -9001
        database_content = self.read_authorizationdb()
        if search_pattern in database_content:
            _updated = '\n'.join((ln for ln in database_content.splitlines() if search_pattern not in ln))
            result = self.update_authorizationdb(_updated)
        return result

    def reset_access_to_service(self, service: str, client: str = '') -> int:
        """Reset all decisions for the specified service.
        The `tccutil` manages the privacy database, which stores decisions the user has made
        about whether apps may access personal data. Reset will cause apps to prompt again
        the next time they access the service.

        If a bundle identifier is specified, the service will be reset for that bundle only.
        """
        return self.executor.execute(f'tccutil reset {service} {client}'.rstrip())

    def read_authorizationdb(self, right: str = 'system.login.console') -> str:
        """Read a login authorization database that controls macOS login progress."""
        return self.executor.get_output(f'{self._cli} -q authorizationdb read {right}')

    def update_authorizationdb(self, payload: str) -> int:
        """Update contents of the login authorization database."""
        return self.executor.sudo(f"-v;sudo {self._cli} -q authorizationdb write system.login.console <<< '{payload}'")



from time import sleep
from typing import Optional

import Quartz

from macuitest.lib.applescript_lib.applescript_wrapper import as_wrapper
from macuitest.lib.operating_system.env import env


class LoginWindow:
    def __init__(self, executor):
        self.executor = executor

    @staticmethod
    def lock_screen():
        if env.version > (10, 13):
            as_wrapper.send_keystroke('q', 'command', 'control')
        else:
            raise NotImplemented

    def unlock_screen(self, password: Optional[str] = None):
        if not self.is_screen_locked:
            return
        self.wake_up()
        as_wrapper.send_keystroke('a', 'command')
        sleep(.5)
        as_wrapper.send_keycode(as_wrapper.key_codes['delete'])
        sleep(.5)
        as_wrapper.typewrite(password or self.executor.get_admin_password())
        sleep(.5)
        self._confirm_login()

    def wake_up(self):
        self.executor.execute('caffeinate -u -t 2')
        for _ in range(3):
            as_wrapper.send_keycode(as_wrapper.key_codes['left'])
            sleep(.5)
        sleep(1)

    @staticmethod
    def _confirm_login():
        as_wrapper.send_keycode(as_wrapper.key_codes['return'])
        sleep(1)

    @property
    def is_screen_locked(self):
        return dict(Quartz.CGSessionCopyCurrentDictionary()).get('CGSSessionScreenIsLocked') is not None



import logging
import os
import subprocess
from typing import Optional, Union


class ShellExecutor:
    execution_timeout_code: int = -1001
    pipe: int = -1
    stdout: int = -2
    devnull: int = -3

    def __init__(self):
        self.__password: Optional[str] = None

    def sudo_get_output(self, cmd: str, timeout: Optional[int] = None) -> Optional[str]:
        """Execute shell instruction with elevated privileges and return execution output."""
        response = self.__execute(cmd=f'echo "{self.get_admin_password()}" | sudo -S -p "" {cmd}', timeout=timeout)
        if isinstance(response, subprocess.CompletedProcess):
            return response.stdout.decode(encoding='utf-8').strip()

    def get_output(self, cmd: str, timeout: Optional[int] = None) -> Optional[str]:
        """Execute shell instruction and return execution output."""
        response = self.__execute(cmd=cmd, timeout=timeout)
        if isinstance(response, subprocess.CompletedProcess):
            return response.stdout.decode(encoding='utf-8').strip()

    def sudo(self, cmd: str, timeout: Optional[int] = None) -> Union[subprocess.CompletedProcess, int]:
        """Execute shell instruction with elevated privileges and return execution code."""
        return self.__execute(cmd=f'echo "{self.get_admin_password()}" | sudo -S -p "" {cmd}', timeout=timeout)

    def execute(self, cmd: str, timeout: Optional[int] = None) -> Union[subprocess.CompletedProcess, int]:
        """Execute a shell instruction and return a `CompletedProcess` object."""
        return self.__execute(cmd=cmd, timeout=timeout)

    def get_admin_password(self, auto_keychain: str = 'com.macuitest.automation'):
        if self.__password is None:
            subprocess.call('sudo -k', shell=True)
            _password = subprocess.getoutput(f'security find-generic-password -a {auto_keychain} -w 2>/dev/null')
            if subprocess.call(f'echo "{_password}" |sudo -S -p "" -v --', shell=True, stderr=-3) == 0:
                self.__password = _password
            else:  # Fallback. Reading environment variables.
                _password = os.environ.get('MACUITEST_PASSWORD')
                if subprocess.call(f'echo "{_password}" |sudo -S -p "" -v --', shell=True, stderr=-3) == 0:
                    self.__password = _password
                else:
                    raise SystemExit(f'\nCANNOT USE SPECIFIED ADMIN PASSWORD:\n'
                                     f'a) Add your actual password to environment variables\n'
                                     f'b) Create a login keychain item with account name "{auto_keychain}" in Keychain Access.app\n'
                                     f'https://support.apple.com/en-gb/guide/keychain-access/kyca1120/mac\n')
        return self.__password

    def __execute(self, cmd: str, timeout: Optional[int] = None) -> Union[subprocess.CompletedProcess, int]:
        """Send instruction to shell."""
        try:
            response = subprocess.run(cmd, shell=True, stderr=self.pipe, stdout=self.pipe, timeout=timeout)
            if response.returncode:
                warning_message = f'Command "{cmd}" failed with exit code {response.returncode}\n\t' \
                                  f'{response.stderr.decode(encoding="utf-8").strip()}'
                logging.debug(warning_message)
            return response
        except subprocess.TimeoutExpired as e:
            logging.error(e)
            return self.execution_timeout_code



"""Wrap objc calls to raise python exception."""

from ApplicationServices import (AXUIElementCopyActionNames, AXUIElementCopyAttributeNames,
                                 AXUIElementCopyAttributeValue, AXUIElementCopyElementAtPosition, AXUIElementGetPid,
                                 AXUIElementIsAttributeSettable, AXUIElementPerformAction, AXUIElementSetAttributeValue,
                                 AXUIElementSetMessagingTimeout, )
from ApplicationServices import (kAXErrorActionUnsupported, kAXErrorAttributeUnsupported, kAXErrorCannotComplete,
                                 kAXErrorFailure, kAXErrorIllegalArgument, kAXErrorInvalidUIElement,
                                 kAXErrorNotImplemented, kAXErrorNoValue, )

from macuitest.lib.elements.native import errors


def get_accessibility_element_attribute(element, attribute):
    """ Return the value of an accessibility object's attribute
            Args:
                element: The AXUIElementRef representing the accessibility object
                attribute: The attribute name

            Returns: the value associated with the specified attribute
    """
    error_code, attr_value = AXUIElementCopyAttributeValue(element, attribute, None)
    error_messages = {
        kAXErrorAttributeUnsupported: "The specified AXUIElementRef does not support the specified attribute.",
        kAXErrorNoValue: "The specified attribute does not have a value.",
        kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.", }
    errors.check_ax_error(error_code, error_messages)
    return attr_value


def check_attribute_settable(element, attribute) -> bool:
    """ Check whether the specified accessibility object's attribute can be modified.
            Args:
                element: The AXUIElementRef representing the accessibility object
                attribute: The attribute name

            Returns: a Boolean value indicating whether the attribute is settable
    """
    error_code, settable = AXUIElementIsAttributeSettable(element, attribute, None)
    error_messages = {
        kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way (often due to a timeout).",
        kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        kAXErrorAttributeUnsupported: "The specified AXUIElementRef does not support the specified attribute.",
        kAXErrorNoValue: "The specified attribute does not have a value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.", }
    errors.check_ax_error(error_code, error_messages)
    return settable


def set_attribute_value(element, attribute, value):
    """ Set the accessibility object's attribute to the specified value
            Args:
                element: The AXUIElementRef representing the accessibility object
                attribute: The attribute name
                value: The new value for the attribute
    """
    error_code = AXUIElementSetAttributeValue(element, attribute, value)
    error_messages = {
        kAXErrorIllegalArgument: "The value is not recognized by the accessible application or one of the other arguments is an illegal value.",
        kAXErrorAttributeUnsupported: "The specified AXUIElementRef does not support the specified attribute.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.", }
    errors.check_ax_error(error_code, error_messages)


def get_element_attribute_names(element):
    """ Get a list of attributes supported by the specified accessibility object
            Args:
                element: The AXUIElementRef representing the accessibility object

            Returns: an array containing the accessibility object's attribute names
    """
    error_code, names = AXUIElementCopyAttributeNames(element, None)
    error_messages = {
        kAXErrorAttributeUnsupported: "The specified AXUIElementRef does not support the specified attribute.",
        kAXErrorIllegalArgument: "One or both of the arguments is an illegal value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorFailure: "There was a system memory failure.",
        kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.", }
    errors.check_ax_error(error_code, error_messages)
    return names


def get_element_action_names(element):
    """ Get a list of all the actions the specified accessibility object can perform.
            Args:
                element: The AXUIElementRef representing the accessibility object

            Returns: an array of actions the accessibility object can perform
                (empty if the accessibility object supports no actions)
    """
    error_code, names = AXUIElementCopyActionNames(element, None)
    error_messages = {kAXErrorIllegalArgument: "One or both of the arguments is an illegal value.",
                      kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
                      kAXErrorFailure: "There was some sort of system memory failure.",
                      kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
                      kAXErrorNotImplemented: "The process does not fully support the accessibility API.", }
    errors.check_ax_error(error_code, error_messages)
    return names


def perform_action_on_element(element, action):
    """ Make the accessibility object perform the specified action.
            Args:
                element: The AXUIElementRef representing the accessibility object
                action: The action to be performed
    """
    error_code = AXUIElementPerformAction(element, action)
    error_messages = {
        kAXErrorActionUnsupported: "The specified AXUIElementRef does not support the specified action (you will also receive this error if you pass in the system-wide accessibility object).",
        kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way or the application has not yet responded.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.", }
    errors.check_ax_error(error_code, error_messages)


def get_accessibility_object_pid(element):
    """ Get the process ID associated with the specified accessibility object.
            Args:
                element: The AXUIElementRef representing an accessibility object

            Returns: the process ID associated with the specified accessibility object
    """
    error_code, pid = AXUIElementGetPid(element, None)
    error_messages = {kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
                      kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.", }
    errors.check_ax_error(error_code, error_messages)
    return pid


def get_accessibility_object_on_screen_position(application, x, y):
    """ Get the accessibility object at the specified position in top-left relative screen coordinates
            Args:
                application: The AXUIElementRef representing the application that
                    contains the screen coordinates (or the system-wide accessibility object)
                x: The horizontal position
                y: The vertical position

            Returns: the accessibility object at the position specified by x and y
    """
    error_code, element = AXUIElementCopyElementAtPosition(application, x, y, None)
    error_messages = {kAXErrorNoValue: "There is no accessibility object at the specified position.",
                      kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
                      kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
                      kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
                      kAXErrorNotImplemented: "The process does not fully support the accessibility API.", }
    errors.check_ax_error(error_code, error_messages)
    return element


def set_accessibility_api_timeout(element, timeout: int):
    """ Set the timeout value used in the accessibility API
            Args:
                element: The AXUIElementRef representing an accessibility object
                timeout: The number of seconds for the new timeout value
    """
    error_code = AXUIElementSetMessagingTimeout(element, timeout)
    error_messages = {
        kAXErrorIllegalArgument: "One or more of the arguments is an illegal value (timeout values must be positive).",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.", }
    errors.check_ax_error(error_code, error_messages)



from distutils.core import setup
from setuptools import find_packages
import subprocess

macos_ver = tuple(int(i) for i in subprocess.getoutput('sw_vers -productVersion').split('.'))


with open('README.md') as fh:
    long_description = fh.read()

setup(
    name='macuitest',
    packages=find_packages(),
    version='0.5.11',
    license='Apache-2.0 License',
    description='A simple UI testing framework for macOS',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Andrii Kislitsyn',
    author_email='andriikislitsyn@gmail.com',
    url='https://github.com/andriykislitsyn',
    download_url='https://github.com/andriykislitsyn/macuitest/archive/v0.5.11-alpha.tar.gz',
    keywords=['Testing', 'UI', 'Functional', 'macOS'],
    install_requires=[
        'biplist',
        'opencv-python == 3.4.8.29' if macos_ver < (10, 13) else 'opencv-python',
        'pillow',
        'pyobjc-framework-ApplicationServices',
        'pyobjc-framework-Cocoa',
        'pyobjc-framework-CoreText',
        'pyobjc-framework-Quartz',
        'PyTweening',
        'webcolors',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Education :: Testing',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



from typing import Optional

import numpy
from Quartz import CoreGraphics as cg, CGDisplayBounds, CGMainDisplayID

from macuitest.config.constants import ScreenSize


class Monitor:
    def __init__(self):
        self.__is_retina: Optional[bool] = None
        self.__screen_size: Optional[ScreenSize] = None

    @property
    def screenshot(self) -> numpy.ndarray:
        image = cg.CGWindowListCreateImage(cg.CGRectInfinite, cg.kCGWindowListOptionOnScreenOnly,
                                           cg.kCGNullWindowID, cg.kCGWindowImageDefault)
        bytes_per_row = cg.CGImageGetBytesPerRow(image) // 4
        image = numpy.frombuffer(cg.CGDataProviderCopyData(cg.CGImageGetDataProvider(image)), dtype=numpy.uint8)
        return image.reshape((self.size.height, bytes_per_row, 4))[:, :self.size.width, :]

    @property
    def bytes(self):
        image = cg.CGWindowListCreateImage(cg.CGRectInfinite, cg.kCGWindowListOptionOnScreenOnly,
                                           cg.kCGNullWindowID, cg.kCGWindowImageDefault)
        return bytes(numpy.frombuffer(cg.CGDataProviderCopyData(cg.CGImageGetDataProvider(image)), dtype=numpy.uint8))

    @property
    def is_retina(self) -> bool:
        if self.__is_retina is None:
            self.__is_retina = self.size.width > 2000
        return self.__is_retina

    @property
    def size(self) -> ScreenSize:
        if self.__screen_size is None:
            size = CGDisplayBounds(CGMainDisplayID()).size
            self.__screen_size = ScreenSize(int(size.width), int(size.height))
        return self.__screen_size


monitor = Monitor()



import os
from typing import Optional, Tuple

from macuitest.lib.operating_system.shell_executor import ShellExecutor


class Env:
    """ macOS environment variables container. """
    _home: Optional[str] = None
    _version: Optional[str] = None

    def __init__(self):
        self.executor = ShellExecutor()

    @property
    def version_str(self) -> str:
        return '.'.join(str(i) for i in env.version).strip()

    @property
    def version_major_str(self) -> str:
        return '.'.join(str(i) for i in self.version_major).strip()

    @property
    def version_major(self) -> Tuple[int, ...]:
        return self.version[:2]

    @property
    def version(self) -> Tuple[int, ...]:
        """Get macOS version."""
        return tuple((int(item) for item in self.__version.split('.')))

    @property
    def __version(self) -> str:
        if self._version is None:
            self._version = self.executor.get_output('sw_vers -productVersion')
        return self._version

    @property
    def applications(self) -> str:
        return '/Applications'

    @property
    def system_applications(self) -> str:
        return f'{self.system}{self.applications}'

    @property
    def temp_dir(self) -> str:
        return os.getenv('TMPDIR')

    @property
    def system(self) -> str:
        return '/System'

    @property
    def system_library(self) -> str:
        return '/Library'

    @property
    def system_app_support(self) -> str:
        return os.path.join(self.system_library, 'Application Support')

    @property
    def system_caches(self) -> str:
        return os.path.join(self.system_library, 'Caches')

    @property
    def system_diag_reports(self) -> str:
        return os.path.join(self.system_logs, 'DiagnosticReports')

    @property
    def system_extensions(self) -> str:
        return os.path.join(self.system_library, 'Extensions')

    @property
    def system_internet_plugins(self) -> str:
        return os.path.join(self.system_library, 'Internet Plug-Ins')

    @property
    def system_launch_daemons(self) -> str:
        return os.path.join(self.system_library, 'LaunchDaemons')

    @property
    def system_logs(self) -> str:
        return os.path.join(self.system_library, 'Logs')

    @property
    def system_preferences(self) -> str:
        return os.path.join(self.system_library, 'Preferences')

    @property
    def system_preference_panes(self) -> str:
        return os.path.join(self.system_library, 'PreferencePanes')

    @property
    def system_privileged_helper_tools(self) -> str:
        return os.path.join(self.system_library, 'PrivilegedHelperTools')

    @property
    def system_security(self) -> str:
        return os.path.join(self.system_library, 'Security')

    @property
    def system_widgets(self) -> str:
        return os.path.join(self.system_library, 'Widgets')

    @property
    def user(self) -> str:
        return os.environ.get('USER')

    @property
    def home(self) -> str:
        return self.__home

    @property
    def trash(self) -> str:
        return os.path.join(self.__home, '.Trash')

    @property
    def desktop(self) -> str:
        return os.path.join(self.__home, 'Desktop')

    @property
    def documents(self) -> str:
        return os.path.join(self.__home, 'Documents')

    @property
    def downloads(self) -> str:
        return os.path.join(self.__home, 'Downloads')

    @property
    def movies(self) -> str:
        return os.path.join(self.__home, 'Movies')

    @property
    def music(self) -> str:
        return os.path.join(self.__home, 'Music')

    @property
    def pictures(self) -> str:
        return os.path.join(self.__home, 'Pictures')

    @property
    def user_caches(self) -> str:
        return os.path.join(self.user_lib, 'Caches')

    @property
    def user_diag_reports(self) -> str:
        return os.path.join(self.user_lib, 'Logs', 'DiagnosticReports')

    @property
    def user_logs(self) -> str:
        return os.path.join(self.user_lib, 'Logs')

    @property
    def user_launch_agents(self) -> str:
        return os.path.join(self.user_lib, 'LaunchAgents')

    @property
    def user_app_support(self) -> str:
        return os.path.join(self.user_lib, 'Application Support')

    @property
    def user_app_script(self) -> str:
        return os.path.join(self.user_lib, 'Application Scripts')

    @property
    def user_saved_app_state(self) -> str:
        return os.path.join(self.user_lib, 'Saved Application State')

    @property
    def user_cookies(self) -> str:
        return os.path.join(self.user_lib, 'Cookies')

    @property
    def user_containers(self) -> str:
        return os.path.join(self.user_lib, 'Containers')

    @property
    def user_group_containers(self) -> str:
        return os.path.join(self.user_lib, 'Group Containers')

    @property
    def user_preferences(self) -> str:
        return os.path.join(self.user_lib, 'Preferences')

    @property
    def user_lib(self) -> str:
        return os.path.join(self.__home, 'Library')

    @property
    def __home(self) -> str:
        """Get path to the user HOME folder."""
        if self._home is None:
            self._home = os.getenv("HOME")
        return self._home


env = Env()



import json
import re
import time
from typing import Tuple, ClassVar
from urllib import error, request


class Networking:
    ports_list: ClassVar[Tuple[str, ...]] = ('en0', 'en1')  # Ethernet port and Wifi port.

    def __init__(self, executor):
        self.executor = executor

    def send_curl_request(self, request_):
        self.executor.execute(f'curl {request_}')

    @property
    def mac_address(self) -> str:
        _device = self.active_internet_device
        _mac_address_raw = self.executor.get_output(f'networksetup -getmacaddress {_device}')
        return _mac_address_raw.replace('Ethernet Address: ', '').replace(f' (Device: {_device})', '').strip()

    @property
    def active_internet_device(self) -> str:
        for device in self.network_service_order:
            _status = self.executor.get_output(f'ifconfig {device} |grep status').split(': ')[-1]
            if _status == 'active':
                return device

    @property
    def network_service_order(self):
        _network_service_order = self.executor.get_output('networksetup listnetworkserviceorder')
        _devices = re.findall(r'Device: \w+-\w+|Device: \w+', _network_service_order)
        _devices = [d.split(': ')[-1] for d in _devices if 'en' in d]
        return _devices

    def turn_internet_on(self):
        """Wait until the Internet connection is established."""
        for port in self.ports_list:
            self.executor.sudo(f'ifconfig {port} up')
        delta_time = time.time() + 30
        while time.time() < delta_time:
            try:
                resp = request.urlopen('https://apple.com')
                if resp.code != 404:
                    return
            except error.URLError:
                time.sleep(0.5)

    def turn_internet_off(self):
        for port in self.ports_list:
            self.executor.sudo(f'ifconfig {port} down')

    @property
    def ip_info(self) -> dict:
        """Get external ip information from: ipinfo.io
        :return dict: {'ip': ..., 'hostname': ..., 'city': ..., 'region': ..., 'country': ..., 'loc': ..., 'org': ...'}
        """
        for _ in range(5):
            try:
                return json.load(request.urlopen('https://ipinfo.io/json'))
            except error.URLError:
                continue
        return dict()



import glob
import logging
import os
import time
from typing import List

from macuitest.lib.applescript_lib.applescript_wrapper import as_wrapper, AppleScriptError
from macuitest.lib.apps.application import Application
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.native.native_ui_element import NativeUIElement
from macuitest.lib.operating_system.env import env
from macuitest.lib.operating_system.macos import macos


class DownloadHasNotStartedException(Exception):
    pass


class DownloadHasNotFinishedException(Exception):
    pass


class MaliciousFileWarning(Exception):
    """Thrown when Chrome shows a warning after downloading a malicious file."""
    pass


class _BaseBrowser(Application):
    def __init__(self, app_name):
        super().__init__(app_name)
        self.browser_name = app_name
        self.document_type = None
        self.download_temp_pattern = None
        self.downloads_count = -1

    def close_tabs(self):
        try:
            as_wrapper.execute(f'tell every window of application "{self.name}" to close tabs')
            wait_condition(lambda: not self._get_tabs_number())
        except AppleScriptError:
            pass

    def _get_tabs_number(self) -> List[List[str]]:
        try:
            return as_wrapper.execute(f'tell application "{self.name}" to get name of every tab of every window')
        except AppleScriptError:
            pass

    def download_file(self, url, file_name='file_pattern*', expected_file_size=2 * 10 ** 5):
        self.launch()
        self.open_link_as_user(url, is_download_confirmation_needed=True)
        self.wait_download_finished(env.downloads)
        self.close_tabs()
        file_ = wait_condition(glob.glob, 30, os.path.join(env.downloads, file_name))
        assert file_, f'File missing. Contents of `Downloads` folder: {os.listdir(env.downloads)}'
        logging.info(f'File: {file_[0]}')
        file_size = os.path.getsize(file_[0])
        logging.info(f'Size: {file_size}')
        assert file_size >= expected_file_size, f'Package size < 200KB. Actual size is: {file_size / 10 ** 3}KB'
        return file_[0]

    @staticmethod
    def open_link_in_a_new_tab(url):
        as_wrapper.send_keystroke('t', 'command')
        time.sleep(1)
        as_wrapper.typewrite(url)
        time.sleep(1)
        as_wrapper.send_keycode(as_wrapper.key_codes['return'])
        time.sleep(.5)

    def open_link_as_user(self, url, is_download_confirmation_needed=False):
        self.open_new_tab()
        self.set_address_bar_value(url)
        as_wrapper.send_keycode(as_wrapper.key_codes['return'])
        if is_download_confirmation_needed:
            self.downloads_count = self.files_in_downloads_dir
            self.confirm_download()

    def open_new_tab(self):
        as_wrapper.tell_app(self.name, f'make new {self.document_type}')
        time.sleep(1)

    def open_urls(self, urls: List[str]):
        for url in urls:
            self.open_link_as_user(url)

    def set_address_bar_value(self, new_value):
        macos.shell_executor.execute(f'printf "{new_value}" |pbcopy')
        as_wrapper.send_keystroke('v', 'command')  # Paste the data
        time.sleep(1)

    def confirm_download(self):
        pass

    def wait_download_finished(self, target_directory, timeout=180):
        """Wait for a temporary file to be removed from the download directory."""
        time.sleep(3)
        if not wait_condition(lambda: self.files_in_downloads_dir > self.downloads_count, 120):
            raise DownloadHasNotStartedException  # Sometimes resolver works slower than we need. So we wait 2m.
        if not wait_condition(lambda: glob.glob(os.path.join(target_directory, self.download_temp_pattern)) == [],
                              timeout):
            raise DownloadHasNotFinishedException

    def webpage_did_load(self, webpage_address):
        assert wait_condition(lambda: isinstance(self.active_tab_url, str), timeout=30)
        return wait_condition(lambda: webpage_address in self.active_tab_url, timeout=30)

    @property
    def active_tab_url(self):
        return wait_condition(lambda: as_wrapper.execute(self._get_url_command), timeout=30)

    @property
    def files_in_downloads_dir(self):
        return len([file_ for file_ in os.listdir(env.downloads) if not file_.startswith('.')])

    @property
    def _get_url_command(self):
        return f'''
            try
                tell application "{self.name}" to set page_address to URL of last tab of window 1
                    if page_address as string = "missing value" then
                        return false
                    end if
                return page_address as string
            end try'''


class _Safari(_BaseBrowser):
    def __init__(self, app_name='Safari'):
        super().__init__(app_name)
        self.document_type = 'document'
        self.download_temp_pattern = '*.download'

    def set_address_bar_value(self, new_value: str):
        self.window.wait_displayed(timeout=10)
        address_bar = NativeUIElement.from_bundle_id('com.apple.Safari').windows()[0].find_element(
            AXIdentifier='WEB_BROWSER_ADDRESS_AND_SEARCH_FIELD', recursive=True)
        address_bar.AXValue = new_value
        time.sleep(1)


safari = _Safari()



"""Convert from common Python types to Apple Event Manager types and vice-versa."""

import datetime
import struct

from Foundation import NSAppleEventDescriptor, NSURL

from macuitest.lib.applescript_lib import aeobjects

__all__ = ['AEConverter', 'AEType', 'AEEnum', 'four_characters_code']


def four_characters_code(code):
    """ Convert four-char code for use in NSAppleEventDescriptor methods.

        code : bytes -- four-char code, e.g. b'utxt'
        Result : int -- OSType, e.g. 1970567284
    """
    return struct.unpack('>I', code)[0]


class AEConverter:
    """ Implements mappings for common Python types with direct AppleScript equivalents. Used by AppleScript class. """

    kMacEpoch = datetime.datetime(1904, 1, 1)
    kUSRF = four_characters_code(aeobjects.keyASUserRecordFields)

    def __init__(self):
        # Clients may add/remove/replace encoder and decoder items:
        self.encoders = {
            NSAppleEventDescriptor.class__(): self.pack_description,
            type(None): self.pack_none,
            bool: self.pack_bool,
            int: self.pack_int,
            float: self.pack_float,
            bytes: self.pack_bytes,
            str: self.pack_str,
            list: self.pack_list,
            tuple: self.pack_list,
            dict: self.pack_dict,
            datetime.datetime: self.pack_datetime,
            AEType: self.pack_type,
            AEEnum: self.pack_enum,
        }

        self.decoders = {four_characters_code(k): v for k, v in {
            aeobjects.typeNull: self.unpack_null,
            aeobjects.typeBoolean: self.unpack_boolean,
            aeobjects.typeFalse: self.unpack_boolean,
            aeobjects.typeTrue: self.unpack_boolean,
            aeobjects.typeSInt32: self.unpack_s_int32,
            aeobjects.typeIEEE64BitFloatingPoint: self.unpack_float64,
            aeobjects.typeUTF8Text: self.unpack_unicode_text,
            aeobjects.typeUTF16ExternalRepresentation: self.unpack_unicode_text,
            aeobjects.typeUnicodeText: self.unpack_unicode_text,
            aeobjects.typeLongDateTime: self.unpack_long_datetime,
            aeobjects.typeAEList: self.unpack_ae_list,
            aeobjects.typeAERecord: self.unpack_ae_record,
            aeobjects.typeAlias: self.unpack_file,
            aeobjects.typeFSS: self.unpack_file,
            aeobjects.typeFSRef: self.unpack_file,
            aeobjects.typeFileURL: self.unpack_file,
            aeobjects.typeType: self.unpack_type,
            aeobjects.typeEnumeration: self.unpack_enumeration,
        }.items()}

    def pack(self, data):
        """Pack Python data.
            data : anything -- a Python value
            Result : NSAppleEventDescriptor -- an AE descriptor, or error if no encoder exists for this type_ of data
        """
        try:
            return self.encoders[data.__class__](data)  # quick lookup by type_/class
        except (KeyError, AttributeError):
            for type_, encoder in self.encoders.items():
                # slower but more thorough lookup that can handle subtypes/subclasses
                if isinstance(data, type_):
                    return encoder(data)
        raise TypeError("Can't pack data into an AEDesc (unsupported type_): {!r}".format(data))

    def unpack(self, desc):
        """Unpack an Apple event descriptor.
            desc : NSAppleEventDescriptor
            Result : anything -- a Python value, or the original NSAppleEventDescriptor if no decoder is found
        """
        decoder = self.decoders.get(desc.descriptorType())
        # unpack known type
        if decoder:
            return decoder(desc)
        # if it's a record-like desc, unpack as dict with an extra AEType(b'pcls') key containing the desc type
        rec = desc.coerceToDescriptorType_(four_characters_code(aeobjects.typeAERecord))
        if rec:
            rec = self.unpack_ae_record(rec)
            rec[AEType(aeobjects.pClass)] = AEType(struct.pack('>I', desc.descriptorType()))
            return rec
        # return as-is
        return desc

    @staticmethod
    def _pack_bytes(description_type, data):
        return NSAppleEventDescriptor.descriptorWithDescriptorType_bytes_length_(
            four_characters_code(description_type), data, len(data))

    @staticmethod
    def pack_description(val):
        return val

    @staticmethod
    def pack_none():
        return NSAppleEventDescriptor.nullDescriptor()

    @staticmethod
    def pack_bool(val):
        return NSAppleEventDescriptor.descriptorWithBoolean_(int(val))

    def pack_int(self, val):
        if (-2 ** 31) <= val < (2 ** 31):
            return NSAppleEventDescriptor.descriptorWithInt32_(val)
        else:
            return self.pack(float(val))

    def pack_float(self, val):
        return self._pack_bytes(aeobjects.typeFloat, struct.pack('d', val))

    def pack_bytes(self, val):
        return self._pack_bytes(aeobjects.typeData, val)

    @staticmethod
    def pack_str(val):
        return NSAppleEventDescriptor.descriptorWithString_(val)

    def pack_datetime(self, val):
        delta = val - self.kMacEpoch
        sec = delta.days * 3600 * 24 + delta.seconds
        return self._pack_bytes(aeobjects.typeLongDateTime, struct.pack('q', sec))

    def pack_list(self, val):
        lst = NSAppleEventDescriptor.listDescriptor()
        for item in val:
            lst.insertDescriptor_atIndex_(self.pack(item), 0)
        return lst

    def pack_dict(self, val):
        record = NSAppleEventDescriptor.recordDescriptor()
        usrf = desctype = None
        for key, value in val.items():
            if isinstance(key, AEType):
                if key.code == aeobjects.pClass and isinstance(value, AEType):
                    # AS packs records that contain a 'class' property by coercing the packed record
                    # to the descriptor type specified by the property's value (assuming it's an AEType)
                    desctype = value
                else:
                    record.setDescriptor_forKeyword_(self.pack(value), four_characters_code(key.code))
            else:
                if not usrf:
                    usrf = NSAppleEventDescriptor.listDescriptor()
                usrf.insertDescriptor_atIndex_(self.pack(key), 0)
                usrf.insertDescriptor_atIndex_(self.pack(value), 0)
        if usrf:
            record.setDescriptor_forKeyword_(usrf, self.kUSRF)
        if desctype:
            new_record = record.coerceToDescriptorType_(four_characters_code(desctype.code))
            if new_record:
                record = new_record
            else:  # coercion failed for some reason, so pack as normal key-value pair
                # noinspection PyUnboundLocalVariable
                record.setDescriptor_forKeyword_(self.pack(desctype), four_characters_code(key.code))
        return record

    @staticmethod
    def pack_type(val):
        return NSAppleEventDescriptor.descriptorWithTypeCode_(four_characters_code(val.code))

    @staticmethod
    def pack_enum(val):
        return NSAppleEventDescriptor.descriptorWithEnumCode_(four_characters_code(val.code))

    @staticmethod
    def unpack_null(desc):
        _ = desc
        return None

    @staticmethod
    def unpack_boolean(desc):
        return desc.booleanValue()

    @staticmethod
    def unpack_s_int32(desc):
        return desc.int32Value()

    @staticmethod
    def unpack_float64(desc):
        return struct.unpack('d', bytes(desc.data()))[0]

    @staticmethod
    def unpack_unicode_text(desc):
        return desc.stringValue()

    def unpack_long_datetime(self, desc):
        return self.kMacEpoch + datetime.timedelta(seconds=struct.unpack('q', bytes(desc.data()))[0])

    def unpack_ae_list(self, desc):
        return [self.unpack(desc.descriptorAtIndex_(i + 1)) for i in range(desc.numberOfItems())]

    def unpack_ae_record(self, desc):
        dct = {}
        for i in range(desc.numberOfItems()):
            key = desc.keywordForDescriptorAtIndex_(i + 1)
            value = desc.descriptorForKeyword_(key)
            if key == self.kUSRF:
                lst = self.unpack_ae_list(value)
                # noinspection PyAssignmentToLoopOrWithParameter
                for i in range(0, len(lst), 2):
                    dct[lst[i]] = lst[i + 1]
            else:
                dct[AEType(struct.pack('>I', key))] = self.unpack(value)
        return dct

    @staticmethod
    def unpack_type(desc):
        return AEType(struct.pack('>I', desc.typeCodeValue()))

    @staticmethod
    def unpack_enumeration(desc):
        return AEEnum(struct.pack('>I', desc.enumCodeValue()))

    @staticmethod
    def unpack_file(desc):
        url = bytes(desc.coerceToDescriptorType_(four_characters_code(aeobjects.typeFileURL)).data()).decode('utf8')
        return NSURL.URLWithString_(url).path()


class AETypeBase:
    """
        Base class for AEType and AEEnum.
        Hashable and comparable, so may be used as keys in dictionaries that map to AE records.
    """
    def __init__(self, code: bytes):
        if not isinstance(code, bytes):
            raise TypeError('invalid code (not a bytes object): {!r}'.format(code))
        elif len(code) != 4:
            raise ValueError('invalid code (not four bytes long): {!r}'.format(code))
        self._code = code

    code = property(lambda self: self._code, doc="bytes -- four-char code, e.g. b'utxt'")

    def __hash__(self):
        return hash(self._code)

    def __eq__(self, val):
        return val.__class__ == self.__class__ and val.code == self._code

    def __ne__(self, val):
        return not self == val

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self._code)


class AEType(AETypeBase):
    """An AE type. Maps to an AppleScript type class, e.g. AEType(b'utxt') <=> 'unicode text'."""


class AEEnum(AETypeBase):
    """An AE enumeration. Maps to an AppleScript constant, e.g. AEEnum(b'yes ') <=> 'yes'."""
